"""Event model run composer from files."""
import time
import typing as tp
import uuid

import numpy as np
from event_model import compose_run, ComposeDescriptorBundle


def gen_stream(
    data_lst: tp.List[dict],
    metadata: dict,
    uid: str = None
) -> tp.Generator[tp.Tuple[str, dict], None, None]:
    """Generate a fake doc stream from data and metadata."""
    crb = compose_run(metadata=metadata, uid=uid if uid else str(uuid.uuid4()))
    yield "start", crb.start_doc
    if len(data_lst) == 0:
        yield "stop", crb.compose_stop()
    else:
        cdb: ComposeDescriptorBundle = crb.compose_descriptor(
            name="primary",
            data_keys=compose_data_keys(data_lst[0])
        )
        yield "descriptor", cdb.descriptor_doc
        for data in data_lst:
            yield "event", cdb.compose_event(data=data, timestamps=compose_timestamps(data))
        yield "stop", crb.compose_stop()


def compose_data_keys(data: tp.Dict[str, tp.Any]) -> tp.Dict[str, dict]:
    """Compose the data keys."""
    return {k: dict(**compose_data_info(v), source="PV:{}".format(k.upper())) for k, v in data.items()}


def compose_data_info(value: tp.Any) -> dict:
    """Compose the data information."""
    if isinstance(value, str):
        return {"dtype": "string", "shape": []}
    elif isinstance(value, float):
        return {"dtype": "number", "shape": []}
    elif isinstance(value, bool):
        return {"dtype": "boolean", "shape": []}
    elif isinstance(value, int):
        return {"dtype": "integer", "shape": []}
    else:
        return {"dtype": "array", "shape": np.shape(value)}


def compose_timestamps(data: tp.Dict[str, tp.Any]) -> tp.Dict[str, float]:
    """Compose the fake time for the data measurement."""
    return {k: time.time() for k in data.keys()}



import typing
from collections import ChainMap

import event_model as em
import numpy as np
import pandas as pd
from bluesky.callbacks.stream import LiveDispatcher
from databroker import Broker, Header
from trackpy import link, locate

from tomography._vend import gen_stream


def gen_processed_images(images: typing.Iterable[typing.Union[list, np.ndarray]],
                         subtrahend: typing.Union[list, np.ndarray]) -> typing.Generator[np.ndarray, None, None]:
    """Generate processed image from a series of images.

    The process procedure is (1) turn ot numpy array, (2) average the frames to a two dimensional image, (3)
    subtract the image and fill zero in negative pixels.

    Parameters
    ----------
    images :
        A iterable of images. The dimensions of each image is no less than 2.
    subtrahend :
        The subtrahend image. The dimensions of each image is no less than 2.

    Yields
    ------
    processed_image :
        A two dimensional image.
    """
    subtrahend = np.asarray(subtrahend)
    subtrahend = get_mean_frame(subtrahend)
    for image in images:
        image = np.asarray(image)
        image = get_mean_frame(image)
        image = subtract_image(image, subtrahend)
        yield image


def get_mean_frame(frames: np.ndarray) -> np.ndarray:
    """Average the frames to a two dimensional image."""
    n = np.ndim(frames)
    if n < 2:
        raise ValueError("The dimension of {} < 2.".format(n))
    elif n == 2:
        mean_frame = np.copy(frames)
    elif n == 3:
        mean_frame = np.mean(frames, axis=0)
    else:
        mean_frame = np.mean(frames, axis=tuple((i for i in range(n - 2))))
    return mean_frame


def subtract_image(minuend: np.ndarray, subtrahend: np.ndarray) -> np.ndarray:
    """Subtract the image and fill zero in negative pixels."""
    ans = np.zeros_like(minuend)
    np.subtract(minuend, subtrahend, out=ans, where=ans > 0)
    return ans


def get_dataframe(run: Header) -> pd.DataFrame:
    """Get the dataframe from the stream. Drop the time column."""
    return run.table().drop(columns=["time"])


class ImageProcessor(LiveDispatcher):
    """A callback to average frames of images, subtract it by another image, and emit the document."""

    def __init__(self, data_key: str, subtrahend: np.ndarray):
        """Initiate the instance.

        Parameters
        ----------
        data_key :
            The key of the data to use.
        subtrahend :
            The 2d image as a subtrahend.
        """
        super(ImageProcessor, self).__init__()
        self.data_key = data_key
        self.subtrahend = np.asarray(subtrahend)

    def start(self, doc, _md=None):
        if _md is None:
            _md = {}
        _md = ChainMap({"analysis_stage": ImageProcessor.__name__}, _md)
        super(ImageProcessor, self).start(doc, _md=_md)

    def event_page(self, doc):
        for event_doc in em.unpack_event_page(doc):
            self.event(event_doc)

    def event(self, doc, **kwargs):
        frames = np.asarray(doc["data"][self.data_key])
        minuend = get_mean_frame(frames)
        result = subtract_image(minuend, self.subtrahend)
        new_data = {k: v for k, v in doc["data"].items() if k != self.data_key}
        new_data[self.data_key] = result.tolist()
        self.process_event({'data': new_data, 'descriptor': doc["descriptor"]})


class PeakTracker(LiveDispatcher):
    """Track the peaks on a series of images and summarize their position and intensity in a dataframe."""

    def __init__(self, data_key: str, diameter: typing.Union[int, tuple], **kwargs):
        """Initiate the instance.

        Parameters
        ----------
        data_key :
            The key of the data to use.
        diameter :
            The pixel size of the peak.
        kwargs :
            The other kwargs for the `trackpy.locate`.
        """
        kwargs["diameter"] = diameter
        super(PeakTracker, self).__init__()
        self.data_key = data_key
        self.config = kwargs

    def start(self, doc, _md=None):
        _md = {"analysis_stage": PeakTracker.__name__}
        super(PeakTracker, self).start(doc, _md=_md)

    def event_page(self, doc):
        for event_doc in em.unpack_event_page(doc):
            self.event(event_doc)

    def event(self, doc, **kwargs):
        image = doc["data"][self.data_key]
        df = locate(image, **self.config)
        df = df.assign(frame=doc["seq_num"])
        for data in df.to_dict("records"):
            self.process_event({"data": data, "descriptor": doc["descriptor"]})


class TrackLinker(LiveDispatcher):
    """Track the peaks in frame and link them in trajectories.

    When a stop is received, the data will be pulled from the databroker and processed. Then, the dataframe will
    be emitted row by row.
    """

    def __init__(self, *, db: Broker = None, search_range: typing.Union[float, tuple], **kwargs):
        """Create the instance.

        Parameters
        ----------
        db :
            The databroker. If None, this callback does nothing.
        search_range :
            The search_range in `trackpy.link`.
        kwargs :
            Other kwargs in `trackpy.link`.
        """
        super(TrackLinker, self).__init__()
        kwargs["search_range"] = search_range
        self.config = kwargs
        self.db = db

    def start(self, doc, _md=None):
        _md = {"analysis_stage": TrackLinker.__name__}
        super(TrackLinker, self).start(doc, _md=_md)

    def event_page(self, doc):
        return

    def event(self, doc, **kwargs):
        return

    def stop(self, doc, _md=None):
        features = get_dataframe(self.db[doc["run_start"]])
        df = link(features, **self.config)
        descriptor = next(iter(self.raw_descriptors.keys()))
        for data in df.to_dict("records"):
            self.process_event({"data": data, "descriptor": descriptor})
        super(TrackLinker, self).stop(doc, _md=None)


class DataFrameDumper(object):
    """Dump the dataframe to the database using databroker."""

    def __init__(self, db: Broker):
        """Create an instance"""
        super(DataFrameDumper, self).__init__()
        self._db = db

    def dump_df(self, df: pd.DataFrame, metadata: dict = None):
        """Dump the data frame into the database with the metadata."""
        if not metadata:
            metadata = {}
        data = df.to_dict("records")
        for name, doc in gen_stream(data, metadata):
            self._db.insert(name, doc)



import typing
from configparser import ConfigParser
from datetime import datetime
from pathlib import Path

import event_model as em
import fire
import numpy as np
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.callbacks.broker import LiveImage
from bluesky.callbacks.core import CallbackBase
from bluesky.callbacks.zmq import RemoteDispatcher
from bluesky.utils import install_kicker
from databroker.core import discover_handlers
from tifffile import TiffWriter

_HANDLERS = discover_handlers()


def myprint(*args, **kwargs) -> None:
    now = datetime.now()
    now_str = now.strftime("%d/%m/%Y %H:%M:%S")
    print("[{}]".format(now_str), *args, **kwargs)
    return


class ServerConfig:

    def __init__(self):
        self.parser = ConfigParser()
        self.parser.add_section("SERVER")
        section = self.parser["SERVER"]
        section["host"] = "localhost"
        section["port"] = "5568"
        section["prefix"] = "raw"
        section["verbose"] = "1"
        section["name"] = "unnamed"

    @property
    def address(self) -> typing.Tuple[str, int]:
        return self.parser.get("SERVER", "host"), self.parser.getint("SERVER", "port")

    @property
    def prefix(self) -> typing.ByteString:
        return self.parser.get("SERVER", "prefix").encode()

    @property
    def verbose(self) -> int:
        return self.parser.getint("SERVER", "verbose")

    @property
    def name(self) -> str:
        return self.parser.get("SERVER", "name")

    @name.setter
    def name(self, value: str) -> None:
        self.parser.set("SERVER", "name", value)
        return

    def read(self, cfg_file: str) -> None:
        path = Path(cfg_file).expanduser()
        if not path.is_file():
            raise FileNotFoundError("No such a file '{}'.".format(cfg_file))
        self.parser.read(cfg_file)

    def write(self, cfg_file: str) -> None:
        path = Path(cfg_file)
        path = path.expanduser()
        if path.is_file():
            raise FileExistsError("File '{}' exists.".format(cfg_file))
        with path.open("w") as f:
            self.parser.write(f)


class ServerBase:

    def __init__(self, config: ServerConfig, callbacks: list):
        self.config = config
        self.callbacks = callbacks
        self.dispatcher = RemoteDispatcher(config.address, prefix=config.prefix)
        router = em.RunRouter([lambda *x: (callbacks, [])], handler_registry=_HANDLERS)
        self.dispatcher.subscribe(router)

    def print(self, *args, **kwargs) -> None:
        if self.config.verbose > 0:
            myprint(*args, **kwargs)

    def run(self) -> None:
        self.print("Start the {} server.".format(self.config.name))
        try:
            self.dispatcher.start()
        except KeyboardInterrupt:
            self.print("Terminate the {} server.".format(self.config.name))


class ExtremumConfig(ServerConfig):

    def __init__(self):
        super().__init__()
        self.start = None
        self.event = None
        self.name = "extremum"
        self.parser.add_section("EXTREMUM")
        section = self.parser["EXTREMUM"]
        section["light_stream"] = "primary"
        section["dark_stream"] = "dark"
        section["data_key"] = "dexela_image"
        section["directory"] = "{start[sample_name]}_{start[uid]}"
        section["file_prefix"] = "{start[sample_name]}_{event[seq_num]}"

    def copy_start(self, start) -> None:
        self.start = dict(start)
        self.start.setdefault("sample_name", "None")

    def set_event(self, event) -> None:
        self.event = event

    @property
    def stream(self) -> str:
        return self.parser.get("EXTREMUM", "light_stream")

    @property
    def dark(self) -> str:
        return self.parser.get("EXTREMUM", "dark_stream")

    @property
    def directory(self) -> Path:
        try:
            return Path(self.parser.get("EXTREMUM", "directory").format(start=self.start))
        except Exception as e:
            print(self.start)
            raise e

    @property
    def max_directory(self) -> Path:
        return self.directory.joinpath("max")

    @property
    def min_directory(self) -> Path:
        return self.directory.joinpath("min")

    @property
    def tiff_directory(self) -> Path:
        return self.directory.joinpath("tiff")

    @property
    def file_prefix(self) -> str:
        return self.parser.get("EXTREMUM", "file_prefix").format(start=self.start, event=self.event)

    @property
    def min_path(self) -> Path:
        rendered = self.file_prefix.format(start=self.start, event=self.event)
        file_name = "{}_min".format(rendered)
        return self.min_directory.joinpath(file_name).with_suffix(".npy")

    @property
    def max_path(self) -> Path:
        rendered = self.file_prefix.format(start=self.start, event=self.event)
        file_name = "{}_max".format(rendered)
        return self.max_directory.joinpath(file_name).with_suffix(".npy")

    @property
    def tiff_path(self) -> Path:
        rendered = self.file_prefix.format(start=self.start, event=self.event)
        file_name = "{}_image".format(rendered)
        return self.tiff_directory.joinpath(file_name).with_suffix(".tiff")

    @property
    def data_key(self) -> str:
        return self.parser.get("EXTREMUM", "data_key")


class Extremum(CallbackBase):

    def __init__(self, config: ExtremumConfig):
        super().__init__()
        self.config: ExtremumConfig = config
        self.min: np.ndarray = None
        self.max: np.ndarray = None
        self.light_uid: str = None
        self.dark_uid: str = None
        self.dark_frame: np.ndarray = None

    def print(self, *args, **kwargs) -> None:
        if self.config.verbose > 0:
            myprint(*args, **kwargs)
        return

    def start(self, doc):
        uid = doc.get("uid", "missing")
        self.print("Start processing (id: '{}').".format(uid))
        self.min = None
        self.max = None
        self.light_uid = None
        self.dark_uid = None
        self.dark_frame = None
        self.config.copy_start(doc)
        self.config.directory.mkdir(parents=True, exist_ok=True)
        self.config.min_directory.mkdir(parents=True, exist_ok=True)
        self.config.max_directory.mkdir(parents=True, exist_ok=True)
        self.config.tiff_directory.mkdir(parents=True, exist_ok=True)
        self.print("Data will be save in {}.".format(str(self.config.directory)))
        return

    def descriptor(self, doc):
        if doc.get("name", "") == self.config.stream:
            self.print("Receive {} stream.".format(self.config.stream))
            self.light_uid = doc["uid"]
        elif doc.get("name", "") == self.config.dark:
            self.print("Receive {} stream.".format(self.config.dark))
            self.dark_uid = doc["uid"]

    def event_page(self, doc):
        for event_doc in em.unpack_event_page(doc):
            self.event(event_doc)
        return

    def event(self, doc):
        self.config.set_event(doc)
        seq_num = doc.get("seq_num", 0)
        frames = doc.get("data", {}).get(self.config.data_key, None)
        if frames is None:
            date_key = self.config.data_key
            self.print("No {} data in the event {}.".format(date_key, seq_num))
            return
        frames_np = np.asarray(frames)
        image = np.mean(frames_np, axis=0, dtype=frames_np.dtype)
        if doc["descriptor"] == self.light_uid:
            # dark subtraction
            self.print("Process light event {}.".format(seq_num))
            image = image - self.dark_frame if self.dark_frame is not None else image
            self.min = np.fmin(self.min, image) if self.min is not None else image
            self.max = np.fmax(self.max, image) if self.max is not None else image
            np.save(str(self.config.min_path), self.min)
            np.save(str(self.config.max_path), self.max)
            tw = TiffWriter(str(self.config.tiff_path))
            tw.write(image)
        elif doc["descriptor"] == self.dark_uid:
            self.print("Process dark event {}.".format(seq_num))
            self.dark_frame = image
        return

    def stop(self, doc):
        uid = doc.get("run_start", "missing")
        self.print("Finish processing (id: '{}').".format(uid))
        return


class ExtremumServer(ServerBase):

    def __init__(self, config: ExtremumConfig):
        extremum = Extremum(config)
        super().__init__(config, [extremum])


class BestEffortConfig(ServerConfig):

    def __init__(self):
        super().__init__()
        self.name = "best_effort"
        section = "BEST EFFORT"
        self.parser.add_section(section)
        self.parser.set(section, "image_key", "dexela_image")

    @property
    def image_key(self):
        return self.parser.get("BEST EFFORT", "image_key")


class BestEffortServer(ServerBase):

    def __init__(self, config: BestEffortConfig):
        bec = BestEffortCallback()
        li = LiveImage(config.image_key)
        super().__init__(config, [bec, li])
        install_kicker(self.dispatcher.loop)


class Commands:
    """A collection of commands.
    """

    def run_extremum(self, cfg_file: str, test: bool = False) -> None:
        """Run extremum server.

        Parameters
        ----------
        cfg_file : str
            The configuration file. It is an .ini file.
        test : bool, optional
            If True, it is a pytest mode, by default False.
        """
        config = ExtremumConfig()
        config.read(cfg_file)
        server = ExtremumServer(config)
        if not test:
            server.run()
        return

    def run_best_effort(self, cfg_file: str, test: bool = False) -> None:
        """Run best effort server.

        Parameters
        ----------
        cfg_file : str
            The configuration file. It is an .ini file.
        test : bool, optional
            If True, it is a pytest mode, by default False.
        """
        config = BestEffortConfig()
        config.read(cfg_file)
        server = BestEffortServer(config)
        if not test:
            server.run()
        return

    def create_extremum_config(self, cfg_file: str) -> None:
        """Create the configuration file for extremum server.

        Parameters
        ----------
        cfg_file : str
            The configuration file. It is an .ini file.
        """
        config = ExtremumConfig()
        config.write(cfg_file)
        return

    def create_best_effort_config(self, cfg_file: str) -> None:
        """Create the configuration file for best efforts server.

        Parameters
        ----------
        cfg_file : str
            The configuration file. It is an .ini file.
        """
        config = BestEffortConfig()
        config.write(cfg_file)
        return


def run_cli():
    c = Commands()
    fire.Fire(
        {
            "run_extremum": c.run_extremum,
            "run_best_effort": c.run_best_effort,
            "create_extremum_config": c.create_extremum_config,
            "create_best_effort_config": c.create_best_effort_config
        }
    )


if __name__ == "__main__":
    run_cli()



import itertools
import math
import time
import typing
import uuid

import bluesky.plan_stubs as bps
import bluesky.plans as bp
import bluesky.preprocessors as bpp
import numpy as np
import ophyd
from bluesky.utils import short_uid
from ophyd import Signal, Kind


class TomoPlanError(Exception):
    """Error in the plans.

    Parameters
    ----------
    Exception : [type]
        [description]
    """
    pass


def _extarct_motor_pos(mtr):
    """Extract the motor position.

    Parameters
    ----------
    mtr : [type]
        [description]

    Returns
    -------
    [type]
        [description]

    Yields
    -------
    [type]
        [description]
    """
    ret = yield from bps.read(mtr)
    if ret is None:
        return None
    return next(
        itertools.chain(
            (ret[k]["value"] for k in mtr.hints.get("fields", [])),
            (v["value"] for v in ret.values()),
        )
    )


def configure_area_det(detector, exposure, acq_time):
    """Configure exposure time of a detector in continuous acquisition mode.

    Parameters
    ----------
    detector : [type]
        [description]
    exposure : [type]
        [description]
    acq_time : [type]
        [description]

    Returns
    -------
    [type]
        [description]

    Yields
    -------
    [type]
        [description]

    Raises
    ------
    TomoPlanError
        [description]
    """
    if exposure < acq_time:
        raise TomoPlanError("exposure time < frame acquisition time: {} < {}".format(exposure, acq_time))
    yield from bps.mv(detector.cam.acquire_time, acq_time)
    res = yield from bps.read(detector.cam.acquire_time)
    real_acq_time = res[detector.cam.acquire_time.name]["value"] if res else 1
    if hasattr(detector, "images_per_set"):
        # compute number of frames
        num_frame = math.ceil(exposure / real_acq_time)
        yield from bps.mv(detector.images_per_set, num_frame)
    else:
        # The dexela detector does not support `images_per_set` so we just
        # use whatever the user asks for as the thing
        num_frame = 1
    computed_exposure = num_frame * real_acq_time
    return num_frame, real_acq_time, computed_exposure


def dark_plan(detector):
    """Take a dark scan in "dark" stream.

    Parameters
    ----------
    detector : [type]
        [description]

    Yields
    -------
    [type]
        [description]
    """
    # Restage to ensure that dark frames goes into a separate file.
    yield from bps.unstage(detector)
    yield from bps.stage(detector)

    yield from bps.trigger_and_read([detector], name="dark")

    # Restage.
    yield from bps.unstage(detector)
    yield from bps.stage(detector)


def _get_motors_and_coords(starts, ends, nums) -> np.ndarray:
    """Get the motors and coordinates of the motors, like [[motor1_pos1, motor2_pos1, ...],
    [motor1_pos2, motor2_pos2, ...], ...].

    Returns
    -------
    [type]
        [description]

    Raises
    ------
    TomoPlanError
        [description]
    TomoPlanError
        [description]
    """
    axes = [np.linspace(start, end, num) for start, end, num in zip(starts, ends, nums)]
    grids = np.meshgrid(*axes, sparse=False, indexing="ij")
    coords = np.column_stack([grid.flatten() for grid in grids])
    return coords


def fly_scan_nd(
    detectors: list,
    *args,
    move_velocity: float,
    time_per_point: float,
    time_per_frame: float,
    shutter: object,
    shutter_open: typing.Any,
    shutter_close: typing.Any,
    shutter_wait_open: float = 0.,
    shutter_wait_close: float = 0.,
    take_dark: bool = True,
    md: dict = None,
    backoff: float = 0.,
    snake: bool = False,
) -> typing.Generic:
    """Move on a grid and do a fly scan at each point in the grid.

    For example, `fly_scan_nd([detector], motor_y, 0, 10, 11,
    motor_x, 0, 20, 21, motor_fly, 0, 5, 6,
    time_per_point=10, time_per_frame=1, shutter=shutter,
    shutter_open=1, shutter_close=0,
    shutter_wait_open=2, shutter_wait_close=5,
    move_velocity=5, take_dark=True,
    md={"task": "fly scan sample 1", backoff=0.5, snake=False})`
    means that

    set detector so that it will collect one image for 10 s
    one image contains 10 frames and each frame for 1 s
    for y in 0, 1, 2, ..., 10:
        for x in 0, 1, 2, ..., 20:
            move to (x, y)
            wait 5 s
            collect dark image during the movement
            open shutter
            wait 2 s
            fly scan the motor_fly from -0.5 to 5.5
            collect 6 images during the fly
            close shutter

    Parameters
    ----------
    detectors : list
        A list of detectors. The first one must be an area detector. The list shouldn't include the motors in the
        `args`.
    *args :
        patterned like (``motor1, start1, stop1, num1,``
                        ``motor2, start2, stop2, num2,``
                        ``motor3, start3, stop3, num3,`` ...
                        ``motorN, startN, stopN, numN``)

        The first motor is the "slowest", the outer loop. The last motor
        is the "fly" motor, the non-stoping scan along an axis. For all motors
        except the first motor, there is a "snake" argument: a boolean
        indicating whether to following snake-like, winding trajectory or a
        simple left-to-right trajectory.
    move_velocity : float
        The speed for the motors to move to the next grid point.
    time_per_point : float
        The time to collect one image at one point.
    time_per_frame : float
        The time to collect one frame in a image. One image contains serveral frames.
    shutter : object
        The fast shutter.
    shutter_open : typing.Any
        The value of the shutter in open state.
    shutter_close : typing.Any
        The value of the shutter in close state.
    shutter_wait_open : float, optional
        The time between the shutter open and the start of the light image collection, by default 0.
    shutter_wait_close : float, optional
        The time between the shutter close and the start of the dark image collection, by default 0.
    take_dark : bool, optional
        If true, take a dark image at the end of the fly scan, by default True
    md : dict, optional
        The dictionary of the metadata to added into the plan, by default None
    backoff : float, optional
        If non-zero, fly scan from start - backoff to end + backoff, by default 0.
    snake : bool, optional
        If true, snake the axis of the fly scan, by default False

    Returns
    -------
    typing.Generic
        The generator of the plan.

    Yields
    -------
    Iterator[typing.Generic]
        The messages of the plan.

    Raises
    ------
    TomoPlanError
        Empty detector list.
    TomoPlanError
        Not enough motors.
    TomoPlanError
        Wrong motor positions format.
    """
    # check args
    if not detectors:
        raise TomoPlanError("dets cannot be an empty list.")
    if len(args) < 8:
        raise TomoPlanError(
            "There must be at least 8 arguments for the motors, like ``motor1, start1, end1, number1, motor2, "
            "start2, end2, number2`.")
    if len(args) % 4 != 0:
        raise TomoPlanError(
            "The arguments must be in format `motor1, start1, end1, number1, motor2, start2, end2, number2, ...`")

    # get the motors and positions
    fly_motor = args[-4]
    fly_start = args[-3]
    fly_stop = args[-2]
    fly_pixels = args[-1]
    motors = args[0:-4:4]
    starts = args[1:-4:4]
    ends = args[2:-4:4]
    nums = args[3:-4:4]
    coords = _get_motors_and_coords(starts, ends, nums)

    # configurea area detector
    ad = detectors[0]
    num_frame, acq_time, computed_dwell_time = yield from configure_area_det(
        ad, time_per_point, time_per_frame
    )

    # set up metadata
    sp = {
        "time_per_frame": acq_time,
        "num_frames": num_frame,
        "requested_exposure": time_per_point,
        "computed_exposure": computed_dwell_time,
        "type": "ct",
        "uid": str(uuid.uuid4()),
        "plan_name": "fly_scan_nd",
    }
    _md = {
        "detectors": [det.name for det in detectors],
        "plan_args": {},
        "map_size": (fly_pixels,) + nums,
        "hints": {},
        "extents": [(fly_start, fly_stop)] + [(start, end) for start, end in zip(starts, ends)],
        **{f"sp_{k}": v for k, v in sp.items()},
    }
    _md.update(md or {})
    _md["hints"].setdefault(
        "dimensions",
        [((f"start_{fly_motor.name}",), "primary")] + [((motor.name,), "primary") for motor in motors],
    )
    # soft signal to use for tracking pixel edges
    px_start = Signal(name=f"start_{fly_motor.name}", kind=Kind.normal)
    px_stop = Signal(name=f"stop_{fly_motor.name}", kind=Kind.normal)
    all_dets = detectors + [px_start, px_stop, fly_motor.velocity] + list(motors)
    # or get the gating working below.
    speed = abs(fly_stop - fly_start) / (fly_pixels * computed_dwell_time)
    # check the speed
    low, high = sorted(fly_motor.velocity.limits)
    if speed < low or speed > high:
        raise ValueError("The fly scan velocity is {}. ".format(speed) +
                         "It is out of the range of the motor speed: ({}, {}).".format(low, high))
    if move_velocity < low or move_velocity > high:
        raise ValueError("The move velocity is {}. ".format(move_velocity) +
                         "It is out of the range of the motor speed: ({}, {}).".format(low, high))

    @bpp.reset_positions_decorator([fly_motor.velocity])
    @bpp.set_run_key_decorator(f"xrd_map_{uuid.uuid4()}")
    @bpp.stage_decorator(detectors)
    @bpp.run_decorator(md=_md)
    def inner():
        _fly_start, _fly_stop = fly_start, fly_stop
        _backoff = backoff
        for coord in coords:
            # move the step motors
            yield from bps.checkpoint()
            yield from bps.mv(fly_motor.velocity, move_velocity)
            pre_fly_group = short_uid("pre_fly")
            for motor, pos in zip(motors, coord):
                yield from bps.abs_set(motor, pos, group=pre_fly_group)
            yield from bps.abs_set(
                fly_motor, _fly_start - _backoff, group=pre_fly_group
            )
            # take a dark during the movement
            if take_dark:
                yield from bps.sleep(shutter_wait_close)
                yield from dark_plan(ad)
            yield from bps.wait(group=pre_fly_group)

            # wait for the pre-fly motion to stop
            yield from bps.mv(fly_motor.velocity, speed)
            yield from bps.mv(shutter, shutter_open)
            yield from bps.sleep(shutter_wait_open)
            fly_group = short_uid("fly")
            yield from bps.abs_set(fly_motor, _fly_stop + _backoff, group=fly_group)
            for j in range(fly_pixels):

                fly_pixel_group = short_uid("fly_pixel")
                for d in detectors:
                    yield from bps.trigger(d, group=fly_pixel_group)

                # grab motor position right after we trigger
                start_pos = yield from _extarct_motor_pos(fly_motor)
                yield from bps.mv(px_start, start_pos)
                # wait for frame to finish
                yield from bps.wait(group=fly_pixel_group)

                # grab the motor position
                stop_pos = yield from _extarct_motor_pos(fly_motor)
                yield from bps.mv(px_stop, stop_pos)
                # generate the event
                yield from bps.create("primary")
                for obj in all_dets:
                    yield from bps.read(obj)
                yield from bps.save()
            yield from bps.checkpoint()
            yield from bps.mv(shutter, shutter_close)
            yield from bps.wait(group=fly_group)
            yield from bps.checkpoint()
            if snake:
                # if snaking, flip these for the next pass through
                _fly_start, _fly_stop = _fly_stop, _fly_start
                _backoff = -_backoff
        return

    return (yield from inner())


def fly_scan_nd_no_shutter(
    detectors: list,
    *args,
    move_velocity: float,
    time_per_point: float,
    time_per_frame: float,
    shutter: object,
    shutter_open: typing.Any,
    shutter_close: typing.Any,
    shutter_wait_open: float = 0.,
    shutter_wait_close: float = 0.,
    take_dark: bool = True,
    md: dict = None,
    backoff: float = 0.,
    snake: bool = False,
) -> typing.Generic:
    """Move on a grid and do a fly scan at each point in the grid.

    For example, `fly_scan_nd([detector], motor_y, 0, 10, 11,
    motor_x, 0, 20, 21, motor_fly, 0, 5, 6,
    time_per_point=10, time_per_frame=1, shutter=shutter,
    shutter_open=1, shutter_close=0,
    shutter_wait_open=2, shutter_wait_close=5,
    move_velocity=5, take_dark=True,
    md={"task": "fly scan sample 1", backoff=0.5, snake=False})`
    means that

    set detector so that it will collect one image for 10 s
    one image contains 10 frames and each frame for 1 s
    for y in 0, 1, 2, ..., 10:
        for x in 0, 1, 2, ..., 20:
            move to (x, y)
            wait 5 s
            collect dark image during the movement
            open shutter
            wait 2 s
            fly scan the motor_fly from -0.5 to 5.5
            collect 6 images during the fly
            close shutter

    Parameters
    ----------
    detectors : list
        A list of detectors. The first one must be an area detector. The list shouldn't include the motors in the
        `args`.
    *args :
        patterned like (``motor1, start1, stop1, num1,``
                        ``motor2, start2, stop2, num2,``
                        ``motor3, start3, stop3, num3,`` ...
                        ``motorN, startN, stopN, numN``)

        The first motor is the "slowest", the outer loop. The last motor
        is the "fly" motor, the non-stoping scan along an axis. For all motors
        except the first motor, there is a "snake" argument: a boolean
        indicating whether to following snake-like, winding trajectory or a
        simple left-to-right trajectory.
    move_velocity : float
        The speed for the motors to move to the next grid point.
    time_per_point : float
        The time to collect one image at one point.
    time_per_frame : float
        The time to collect one frame in a image. One image contains serveral frames.
    shutter : object
        The fast shutter.
    shutter_open : typing.Any
        The value of the shutter in open state.
    shutter_close : typing.Any
        The value of the shutter in close state.
    shutter_wait_open : float, optional
        The time between the shutter open and the start of the light image collection, by default 0.
    shutter_wait_close : float, optional
        The time between the shutter close and the start of the dark image collection, by default 0.
    take_dark : bool, optional
        If true, take a dark image at the end of the fly scan, by default True
    md : dict, optional
        The dictionary of the metadata to added into the plan, by default None
    backoff : float, optional
        If non-zero, fly scan from start - backoff to end + backoff, by default 0.
    snake : bool, optional
        If true, snake the axis of the fly scan, by default False

    Returns
    -------
    typing.Generic
        The generator of the plan.

    Yields
    -------
    Iterator[typing.Generic]
        The messages of the plan.

    Raises
    ------
    TomoPlanError
        Empty detector list.
    TomoPlanError
        Not enough motors.
    TomoPlanError
        Wrong motor positions format.
    """
    # check args
    if not detectors:
        raise TomoPlanError("dets cannot be an empty list.")
    if len(args) < 8:
        raise TomoPlanError(
            "There must be at least 8 arguments for the motors, like ``motor1, start1, end1, number1, motor2, "
            "start2, end2, number2`.")
    if len(args) % 4 != 0:
        raise TomoPlanError(
            "The arguments must be in format `motor1, start1, end1, number1, motor2, start2, end2, number2, ...`")

    # get the motors and positions
    fly_motor = args[-4]
    fly_start = args[-3]
    fly_stop = args[-2]
    fly_pixels = args[-1]
    motors = args[0:-4:4]
    starts = args[1:-4:4]
    ends = args[2:-4:4]
    nums = args[3:-4:4]
    coords = _get_motors_and_coords(starts, ends, nums)

    # configurea area detector
    ad = detectors[0]
    num_frame, acq_time, computed_dwell_time = yield from configure_area_det(
        ad, time_per_point, time_per_frame
    )

    # set up metadata
    sp = {
        "time_per_frame": acq_time,
        "num_frames": num_frame,
        "requested_exposure": time_per_point,
        "computed_exposure": computed_dwell_time,
        "type": "ct",
        "uid": str(uuid.uuid4()),
        "plan_name": "fly_scan_nd",
    }
    _md = {
        "detectors": [det.name for det in detectors],
        "plan_args": {},
        "map_size": (fly_pixels,) + nums,
        "shape": (fly_pixels,) + nums,
        "hints": {},
        "extents": [(fly_start, fly_stop)] + [(start, end) for start, end in zip(starts, ends)],
        **{f"sp_{k}": v for k, v in sp.items()},
    }
    _md.update(md or {})
    _md["hints"].setdefault(
        "dimensions",
        [((f"start_{fly_motor.name}",), "primary")] + [((motor.name,), "primary") for motor in motors],
    )
    # soft signal to use for tracking pixel edges
    px_start = Signal(name=f"start_{fly_motor.name}", kind=Kind.normal)
    px_stop = Signal(name=f"stop_{fly_motor.name}", kind=Kind.normal)
    all_dets = detectors + [px_start, px_stop, fly_motor.velocity] + list(motors)
    # or get the gating working below.
    speed = abs(fly_stop - fly_start) / (fly_pixels * computed_dwell_time)
    # check the speed
    low, high = sorted(fly_motor.velocity.limits)
    if speed < low or speed > high:
        raise ValueError("The fly scan velocity is {}. ".format(speed) +
                         "It is out of the range of the motor speed: ({}, {}).".format(low, high))
    if move_velocity < low or move_velocity > high:
        raise ValueError("The move velocity is {}. ".format(move_velocity) +
                         "It is out of the range of the motor speed: ({}, {}).".format(low, high))

    @bpp.reset_positions_decorator([fly_motor.velocity])
    @bpp.set_run_key_decorator(f"xrd_map_{uuid.uuid4()}")
    @bpp.stage_decorator(detectors)
    @bpp.run_decorator(md=_md)
    def inner():
        _fly_start, _fly_stop = fly_start, fly_stop
        _backoff = backoff
        yield from bps.mv(shutter, shutter_open)
        for coord in coords:
            # move the step motors
            yield from bps.checkpoint()
            yield from bps.mv(fly_motor.velocity, move_velocity)
            pre_fly_group = short_uid("pre_fly")
            for motor, pos in zip(motors, coord):
                yield from bps.abs_set(motor, pos, group=pre_fly_group)
            yield from bps.abs_set(
                fly_motor, _fly_start - _backoff, group=pre_fly_group
            )
            # take a dark during the movement
            if take_dark:
                yield from bps.mv(shutter, shutter_close)
                yield from bps.sleep(shutter_wait_close)
                yield from dark_plan(ad)
                yield from bps.sleep(shutter_wait_open)
                yield from bps.mv(shutter, shutter_open)
            yield from bps.wait(group=pre_fly_group)

            fly_group = short_uid("fly")
            yield from bps.mv(fly_motor.velocity, speed)
            yield from bps.abs_set(fly_motor, _fly_stop + _backoff, group=fly_group)
            for j in range(fly_pixels):

                fly_pixel_group = short_uid("fly_pixel")
                for d in detectors:
                    yield from bps.trigger(d, group=fly_pixel_group)

                # grab motor position right after we trigger
                start_pos = yield from _extarct_motor_pos(fly_motor)
                yield from bps.mv(px_start, start_pos)
                # wait for frame to finish
                yield from bps.wait(group=fly_pixel_group)

                # grab the motor position
                stop_pos = yield from _extarct_motor_pos(fly_motor)
                yield from bps.mv(px_stop, stop_pos)
                # generate the event
                yield from bps.create("primary")
                for obj in all_dets:
                    yield from bps.read(obj)
                yield from bps.save()
            yield from bps.checkpoint()
            yield from bps.wait(group=fly_group)
            yield from bps.checkpoint()
            if snake:
                # if snaking, flip these for the next pass through
                _fly_start, _fly_stop = _fly_stop, _fly_start
                _backoff = -_backoff
        yield from bps.mv(shutter, shutter_close)
        return

    return (yield from inner())


def grid_scan_nd(
    detectors: list,
    *args,
    snake: typing.Union[list, bool] = None,
    time_per_point: float,
    time_per_frame: float,
    shutter: object,
    shutter_open: typing.Any,
    shutter_close: typing.Any,
    shutter_wait_open: float = 0.,
    shutter_wait_close: float = 0.,
    take_dark: bool = True,
    md=None
) -> typing.Generic:
    """Scan over a mesh; each motor is on an independent trajectory.

    Parameters
    ----------
    detectors : list
        A list of 'readable' objects
    *args :
        patterned like (``motor1, start1, stop1, num1,``
                        ``motor2, start2, stop2, num2,``
                        ``motor3, start3, stop3, num3,`` ...
                        ``motorN, startN, stopN, numN``)

        The first motor is the "slowest", the outer loop. The last motor
        is the "fly" motor, the non-stoping scan along an axis. For all motors
        except the first motor, there is a "snake" argument: a boolean
        indicating whether to following snake-like, winding trajectory or a
        simple left-to-right trajectory.
    snake : bool, optional
        If true, snake the axis of the fly scan, by default None
    time_per_point : float
        The time to collect one image at one point.
    time_per_frame : float
        The time to collect one frame in a image. One image contains serveral frames.
    time_per_point : float
        The time to collect one image at one point.
    time_per_frame : float
        The time to collect one frame in a image. One image contains serveral frames.
    shutter : object
        The fast shutter.
    shutter_open : typing.Any
        The value of the shutter in open state.
    shutter_close : typing.Any
        The value of the shutter in close state.
    shutter_wait_open : float, optional
        The time between the shutter open and the start of the light image collection, by default 0.
    shutter_wait_close : float, optional
        The time between the shutter close and the start of the dark image collection, by default 0.
    take_dark : bool, optional
        If true, take a dark image at the end of the fly scan, by default True
    md : [type], optional
        The dictionary of the metadata to added into the plan, by default None, by default None

    Returns
    -------
    typing.Generic
        The generator of the plan.

    Yields
    -------
    Iterator[typing.Generic]
        The messages of the plan.

    Raises
    ------
    TomoPlanError
        Empty detector list.
    TomoPlanError
        Not enough motors.
    TomoPlanError
        Wrong motor positions format.
    """
    if not md:
        md = {}
    if len(args) == 0:
        raise TomoPlanError("Missing arguments for the motors.")
    if len(args) % 4 != 0:
        raise TomoPlanError(
            "The arguments must be in format `motor1, start1, end1, number1, motor2, start2, end2, number2, ...`")
    if len(detectors) == 0:
        raise TomoPlanError("dets cannot be an empty list.")

    slow_motor = args[0]

    def _per_step(detectors, step, pos_cache):
        # start movement
        move_group = short_uid("move")
        for motor, pos in step.items():
            yield from bps.abs_set(motor, pos, group=move_group)
        # take dark if there is a movement of the slow motor
        if step.get(slow_motor) != pos_cache[slow_motor] and take_dark:
            yield from bps.mv(shutter, shutter_close)
            yield from bps.sleep(shutter_wait_close)
            yield from dark_plan(detectors[0])
            yield from bps.mv(shutter, shutter_open)
        # wait until finish and take reading
        yield from bps.wait(move_group)
        yield from bps.trigger_and_read(detectors)
        # update the cache
        for motor, pos in step.items():
            pos_cache[motor] = pos

    num_frame, time_per_frame2, time_per_point2 = yield from configure_area_det(
        detectors[0], time_per_point, time_per_frame
    )
    _md = {
        "sp_time_per_frame": time_per_frame2,
        "sp_num_frames": num_frame,
        "sp_requested_exposure": time_per_point,
        "sp_computed_exposure": time_per_point2,
        "sp_type": "ct",
        "sp_uid": str(uuid.uuid4()),
        "sp_plan_name": "grid_scan_nd",
    }
    _md.update(md)

    plan = bp.grid_scan(detectors, *args, snake_axes=snake, per_step=_per_step, md=_md)
    if not take_dark:
        plan = bpp.pchain(bps.mv(shutter, shutter_open), bps.sleep(shutter_wait_open), plan)
    plan = bpp.finalize_wrapper(plan, bps.mv(shutter, shutter_close))
    return (yield from plan)


def grid_scan_no_dark(
    detectors: list,
    *args,
    snake: typing.Union[list, bool] = None,
    time_per_point: float,
    time_per_frame: float,
    shutter: object,
    shutter_open: typing.Any,
    shutter_close: typing.Any,
    shutter_wait_open: float = 0.,
    shutter_wait_close: float = 0.,
    md=None
) -> typing.Generic:
    """Scan over a mesh; each motor is on an independent trajectory.

    Parameters
    ----------
    detectors : list
        A list of 'readable' objects
    *args :
        patterned like (``motor1, start1, stop1, num1,``
                        ``motor2, start2, stop2, num2,``
                        ``motor3, start3, stop3, num3,`` ...
                        ``motorN, startN, stopN, numN``)

        The first motor is the "slowest", the outer loop. The last motor
        is the "fly" motor, the non-stoping scan along an axis. For all motors
        except the first motor, there is a "snake" argument: a boolean
        indicating whether to following snake-like, winding trajectory or a
        simple left-to-right trajectory.
    snake : bool, optional
        If true, snake the axis of the fly scan, by default None
    time_per_point : float
        The time to collect one image at one point.
    time_per_frame : float
        The time to collect one frame in a image. One image contains serveral frames.
    time_per_point : float
        The time to collect one image at one point.
    time_per_frame : float
        The time to collect one frame in a image. One image contains serveral frames.
    shutter : object
        The fast shutter.
    shutter_open : typing.Any
        The value of the shutter in open state.
    shutter_close : typing.Any
        The value of the shutter in close state.
    shutter_wait_open : float, optional
        The time between the shutter open and the start of the light image collection, by default 0.
    shutter_wait_close : float, optional
        The time between the shutter close and the start of the dark image collection, by default 0.
    take_dark : bool, optional
        If true, take a dark image at the end of the fly scan, by default True
    md : [type], optional
        The dictionary of the metadata to added into the plan, by default None, by default None

    Returns
    -------
    typing.Generic
        The generator of the plan.

    Yields
    -------
    Iterator[typing.Generic]
        The messages of the plan.

    Raises
    ------
    TomoPlanError
        Empty detector list.
    TomoPlanError
        Not enough motors.
    TomoPlanError
        Wrong motor positions format.
    """
    if not md:
        md = {}
    if len(args) == 0:
        raise TomoPlanError("Missing arguments for the motors.")
    if len(args) % 4 != 0:
        raise TomoPlanError(
            "The arguments must be in format `motor1, start1, end1, number1, motor2, start2, end2, number2, ...`")
    if len(detectors) == 0:
        raise TomoPlanError("dets cannot be an empty list.")

    num_frame, time_per_frame2, time_per_point2 = yield from configure_area_det(
        detectors[0], time_per_point, time_per_frame
    )
    _md = {
        "sp_time_per_frame": time_per_frame2,
        "sp_num_frames": num_frame,
        "sp_requested_exposure": time_per_point,
        "sp_computed_exposure": time_per_point2,
        "sp_type": "ct",
        "sp_uid": str(uuid.uuid4()),
        "sp_plan_name": "grid_scan_nd",
    }
    _md.update(md)

    plan = bp.grid_scan(detectors, *args, snake_axes=snake, md=_md)
    plan = bpp.pchain(bps.mv(shutter, shutter_open), bps.sleep(shutter_wait_open), plan)
    plan = bpp.finalize_wrapper(plan, bps.mv(shutter, shutter_close))
    return (yield from plan)


def loop_until(motor: ophyd.Device, left: float, right: float, t: float) -> typing.Generic:
    """Move motor from left to right and right to left repeatedly until t seconds pass"""
    t0 = time.time()
    while time.time() - t0 < t:
        yield from bps.mv(motor, left)
        yield from bps.mv(motor, right)
    return



from bluesky import RunEngine
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.simulators import summarize_plan

import tomography.plans as plans
import tomography.sim as sim

_run_RE = False
if _run_RE:
    bec = BestEffortCallback()
    bec.disable_plots()
    bec.disable_baseline()
    RE = RunEngine()
    RE.subscribe(bec, "all")


def test_fly_scan_2d():
    shutter = sim.SynSignal(name="shutter")
    area_det = sim.FakeAreaDetector(name="dexela")
    area_det.cam.configure({"acquire_time": 1.})
    motor_y = sim.DelayedSynAxis(name="motor_y")
    motor_y.configure({"velocity": 1.})
    motor_x = sim.DelayedSynAxis(name="motor_x")
    motor_x.configure({"velocity": 1.})

    plan = plans.fly_scan_nd(
        [area_det], motor_y, 0., 1., 2, motor_x, 0., 2., 3,
        time_per_point=1., time_per_frame=1., move_velocity=1., shutter=shutter, shutter_close=1, shutter_open=0.,
        shutter_wait_close=0., shutter_wait_open=0.
    )

    if _run_RE:
        RE(plan)
    else:
        summarize_plan(plan)


def test_fly_scan_3d():
    shutter = sim.SynSignal(name="shutter")
    area_det = sim.FakeAreaDetector(name="dexela")
    area_det.cam.configure({"acquire_time": 1.})
    motor_y = sim.DelayedSynAxis(name="motor_y")
    motor_y.configure({"velocity": 1.})
    motor_z = sim.DelayedSynAxis(name="motor_z")
    motor_z.configure({"velocity": 1.})
    motor_x = sim.DelayedSynAxis(name="motor_x")
    motor_x.configure({"velocity": 1.})

    plan = plans.fly_scan_nd(
        [area_det], motor_z, 2., 3., 2, motor_y, 1., 2., 2, motor_x, 0., 1., 2,
        time_per_point=1., time_per_frame=1., move_velocity=1., shutter=shutter, shutter_close=1, shutter_open=0.,
        shutter_wait_close=0., shutter_wait_open=0.
    )
    if _run_RE:
        RE(plan)
    else:
        summarize_plan(plan)


def test_grid_scan_2d():
    shutter = sim.SynSignal(name="shutter")
    area_det = sim.FakeAreaDetector(name="dexela")
    area_det.cam.configure({"acquire_time": 1.})
    motor_y = sim.DelayedSynAxis(name="motor_y")
    motor_y.configure({"velocity": 1.})
    motor_x = sim.DelayedSynAxis(name="motor_x")
    motor_x.configure({"velocity": 1.})

    plan = plans.grid_scan_nd(
        [area_det], motor_y, 0., 1., 2, motor_x, 0., 2., 3,
        time_per_point=1., time_per_frame=1., shutter=shutter, shutter_close=1., shutter_open=0.,
        shutter_wait_close=0., shutter_wait_open=0.
    )

    if _run_RE:
        RE(plan)
    else:
        summarize_plan(plan)


def test_loop_until():
    motor_x = sim.DelayedSynAxis(name="motor_x")
    motor_x.configure({"velocity": 10.})
    plans.loop_until(motor_x, 0, 1, 0.2)



from pathlib import Path

import numpy as np
from pdfstream.callbacks.composer import gen_stream

import tomography.servers as servers


def test_Extremum(tmpdir):
    """Test if it computes the correct min and max."""
    base = Path(str(tmpdir))
    # callback
    config = servers.ExtremumConfig()
    data_key = "image"
    config.parser["EXTREMUM"]["directory"] = str(base)
    config.parser["EXTREMUM"]["data_key"] = data_key
    extremum = servers.Extremum(config)
    # create date stream and let it flow through the callback
    shape = (2, 3, 3)
    data = [{data_key: np.zeros(shape)}, {data_key: np.ones(shape)}]
    metadata = {"sample_name": "A"}
    stream = gen_stream(data, metadata)
    for name, doc in stream:
        extremum(name, doc)
    # check the files
    file1 = base.joinpath("min").joinpath("A_1_min.npy")
    file2 = base.joinpath("max").joinpath("A_1_max.npy")
    file3 = base.joinpath("min").joinpath("A_2_min.npy")
    file4 = base.joinpath("max").joinpath("A_2_max.npy")
    for f in (file1, file2, file3, file4):
        assert f.is_file()
    # check the output
    arr1 = np.load(str(file1))
    arr2 = np.load(str(file2))
    arr3 = np.load(str(file3))
    arr4 = np.load(str(file4))
    shape2 = (3, 3)
    assert np.array_equal(arr1, np.zeros(shape2, int))
    assert np.array_equal(arr2, np.zeros(shape2, int))
    assert np.array_equal(arr3, np.zeros(shape2, int))
    assert np.array_equal(arr4, np.ones(shape2, int))


def test_ExtremumServer(tmpdir):
    """Test if it can write and read the cfg file"""
    base = Path(str(tmpdir))
    ss = servers.Commands()
    cfg_file = base.joinpath("test_extremum.ini")
    ss.create_extremum_config(str(cfg_file))
    print(cfg_file.read_text())


def test_BestEffortServer(tmpdir):
    base = Path(str(tmpdir))
    ss = servers.Commands()
    cfg_file = base.joinpath("test_best_effort.ini")
    ss.create_best_effort_config(str(cfg_file))
    print(cfg_file.read_text())



from pprint import pformat

import databroker
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import trackpy as tp
from pdfstream.callbacks.composer import gen_stream
from pkg_resources import resource_filename

import tomography.callbacks as cbs


def print_doc(name, doc):
    print(name, "\n", pformat(doc))


def test_ImageProcessor():
    """Check that ImageProcessor is correctly subtracting images."""
    data_key = "pe1_image"

    def verify(_name, _doc):
        if _name != "event":
            return
        data = _doc["data"][data_key]
        assert isinstance(data, list)
        assert np.array_equal(np.asarray(data), np.zeros((3, 3)))

    ip = cbs.ImageProcessor(data_key=data_key, subtrahend=np.ones((3, 3)))
    ip.subscribe(verify, name="event")

    frames = np.ones((2, 3, 3))
    for name, doc in gen_stream([{data_key: frames}], {}):
        ip(name, doc)


def test_gen_processed_images():
    """Test gen_processed_images."""
    images1 = (np.ones((2, 3, 3)) for _ in range(3))
    subtrahend = np.ones((3, 3))
    subtrahend[0, 0] = 2
    images2 = cbs.gen_processed_images(images1, subtrahend=subtrahend)
    for image in images2:
        assert np.array_equal(image, np.zeros((3, 3)))


def test_PeakTracker(tmpdir):
    """Check that PeakTrack and TrackLinker works without errors."""
    tp.quiet()
    # make images
    image_file = resource_filename("tomography", "data/image.png")
    image = plt.imread(image_file)
    images = [image] * 3
    # check if db friendly
    db = databroker.v1.temp()
    # check features
    data_key = "pe1_image"
    pt = cbs.PeakTracker(data_key=data_key, diameter=(11, 11))
    pt.subscribe(db.insert)
    data = [{data_key: image} for image in images]
    for name, doc in gen_stream(data, {}):
        pt(name, doc)
    df = cbs.get_dataframe(db[-1])
    print(df.to_string())
    # check trajectories
    tl = cbs.TrackLinker(db=db, search_range=3)
    tl.subscribe(db.insert)
    for name, doc in db[-1].documents(fill="yes"):
        tl(name, doc)
    df = cbs.get_dataframe(db[-1])
    print(df.to_string())


def test_DataFrameDumper():
    """Test DataFrameDumper."""
    db = databroker.v1.temp()
    dfd = cbs.DataFrameDumper(db)
    data = [1, 2, 3]
    df = pd.DataFrame({"a": [1, 2, 3]})
    metadata = {"key": "a"}
    dfd.dump_df(df, metadata)
    run = db[-1]
    assert run.start["key"] == metadata["key"]
    assert list(run.data("a")) == data



import numpy as np
from ophyd import Device, Component, Kind
from ophyd.sim import DirectImage, SynSignal, SynAxis


class FakeCam(DirectImage):
    acquire_time = Component(SynSignal, kind=Kind.config)


class FakeAreaDetector(Device):
    cam = Component(FakeCam, func=lambda: np.ones(5, 5), kind=Kind.normal)
    images_per_set = Component(SynSignal, kind=Kind.config)

    def trigger(self):
        self.cam.img.exposure_time = self.cam.acquire_time.value * self.images_per_set.value
        return super(FakeAreaDetector, self).trigger()


class DelayedSynAxis(SynAxis):

    def set(self, value):
        self.delay = abs(self.position - value) / self.velocity.value
        return super(DelayedSynAxis, self).set(value)



import math
import typing

import fabio
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tqdm
import trackpy as tp
import xarray as xr
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from xarray.plot import FacetGrid

_VERBOSE = 1


def set_verbose(level: int) -> None:
    global _VERBOSE
    _VERBOSE = level


def my_print(*args, **kwargs) -> None:
    if _VERBOSE > 0:
        print(*args, **kwargs)


def reshape(dataset: xr.Dataset, name: str, inverted: bool = True) -> xr.DataArray:
    """
    Reshape the xarray dataset[name] into two dimensional array. Return a reshape data array with coordinates.

    Use `shape`, `snaking`, `extents` in the dataset.attrs. The axis axis will be converted to the relative
    position to samples so that the coordinate is the negative motor position.
    """
    reshaped = _reshape(dataset[name].values, dataset.attrs["shape"], dataset.attrs["snaking"])
    coords = _get_coords(dataset.attrs, inverted=inverted)
    return xr.DataArray(reshaped, coords=coords, dims=list(coords.keys()))


def _reshape(arr: np.ndarray, shape: typing.List[int], snaking: typing.List[bool]) -> np.ndarray:
    reshaped = arr.reshape(shape)
    if len(snaking) > 1 and snaking[1]:
        new_reshaped = reshaped.copy()
        for i, row in enumerate(reshaped):
            if i % 2 == 1:
                new_reshaped[i] = row[::-1]
        reshaped = new_reshaped
    return reshaped


def plot_real_aspect(xarr: xr.DataArray, *args, alpha: float = 1.6, **kwargs) -> xr.plot.FacetGrid:
    """Visualize two dimensional arr as a color map. The color ranges from median - alpha * std to median +
    alpha * std."""
    facet = xarr.plot(*args, **kwargs, **get_vlim(xarr, alpha))
    set_real_aspect(facet.axes)
    return facet


def get_vlim(xarr: xr.DataArray, alpha: float) -> dict:
    """Get vmin, vmax using mean and std."""
    mean = xarr.mean()
    std = xarr.std()
    return {"vmin": max(0., mean - alpha * std), "vmax": mean + alpha * std}


def annotate_peaks(df: pd.DataFrame, image: xr.DataArray, ax: plt.Axes = None, alpha: float = 1.6,
                   **kwargs) -> None:
    """A function wrapping the tp.annotate. Use different default setting."""
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    imshow_style = dict(**get_vlim(image, alpha=alpha), cmap="viridis")
    imshow_style.update(kwargs)
    tp.annotate(df, image, ax=ax, imshow_style=imshow_style)
    ax.set_ylim(*ax.get_ylim()[::-1])


def create_atlas(df: pd.DataFrame, start_frame: int = 1, inverted: bool = True,
                 excluded: set = frozenset(["frame"])) -> xr.Dataset:
    """Create the dataset of the maps of grains.

    The dataset is like below.

        Dimensions:   (dim_0: 2, dim_1: 2, grain: 2)
        Coordinates:
          * dim_0     (dim_0) float64 6.0 0.0
          * dim_1     (dim_1) float64 2.0 0.0
          * grain     (grain) int64 0 1
        Data variables:
            maps      (grain, dim_0, dim_1) float64 1.0 0.0 0.0 0.0 0.0 0.0 0.0 2.0
            y         (grain) int64 1 2
            x         (grain) int64 1 2
            mass      (grain) int64 1 2
            size      (grain) int64 1 2
            ecc       (grain) int64 1 2
            signal    (grain) int64 1 2
            raw_mass  (grain) int64 1 2

    Parameters
    ----------
    df :
        The dataframe of trajectories. It has columns "mass" and "frame" and attrs "extents", "shape", "snaking".
    start_frame :
        The starting number of the frame.
    inverted :
        If True, invert all axis so that maps are viewed in sample frame.
    excluded :
        The excluded the columns in dataframe.

    Returns
    -------
    ds :
        A dataset created.
    """
    all_data = dict()
    groups = df.groupby("particle", sort=False)
    # get a stack of maps
    data_list = []
    names = []
    for name, group in groups:
        data = _fill_in_shape(group, start_frame)
        data_list.append(data)
        names.append(name)
    coords = _get_coords(df.attrs, inverted)
    dims = ["grain"] + list(coords.keys())
    all_data["maps"] = (dims, np.stack(data_list))
    del data_list
    # get additional data
    mean_df = groups.mean()
    for key, lst in mean_df.to_dict("list").items():
        if key not in excluded:
            all_data[key] = (["grain"], lst)
    # add grain to coords
    coords["grain"] = names
    return xr.Dataset(all_data, coords=coords)


def _fill_in_shape(df: pd.DataFrame, start_frame: int):
    """Create a map."""
    start_doc = df.attrs
    shape = tuple(start_doc["shape"])
    snaking_tup = tuple(start_doc["snaking"])
    snaking = snaking_tup[1] if len(snaking_tup) > 1 else False
    data = np.zeros(shape)
    for row in df.itertuples():
        seq_num = int(row.frame) - start_frame
        if seq_num < 0:
            raise ValueError(
                "Frame number smaller than the starting number: {}, {}".format(int(row.frame), start_frame))
        pos = _get_pos(seq_num, shape, snaking)
        data[pos] = row.mass
    return data


def _get_pos(seq_num: int, raster_shape: tuple, snaking: bool):
    """Get the index in the matrix."""
    pos = list(np.unravel_index(seq_num, raster_shape))
    if snaking and (pos[0] % 2):
        pos[1] = raster_shape[1] - pos[1] - 1
    return tuple(pos)


def _get_coords(start_doc: dict, inverted: bool) -> dict:
    """Get coordinates."""
    if "shape" not in start_doc:
        raise KeyError("Missing key '{}' in the metadata.".format("shape"))
    if "extents" not in start_doc:
        raise KeyError("Missing key '{}' in the metadata".format("extents"))
    shape = start_doc["shape"]
    extents = [np.asarray(extent) - np.min(extent) for extent in start_doc["extents"]]
    if inverted:
        extents = [extent[::-1] for extent in extents]
    coords = [np.linspace(*extent, num) for extent, num in zip(extents, shape)]
    return {"dim_{}".format(i): data for i, data in enumerate(coords)}


def plot_grain_maps(atlas: xr.Dataset, name: str = "maps", col: str = "grain", **kwargs) -> xr.plot.FacetGrid:
    """Plot the grain maps from the atlas, the output from `create_atlas`."""
    kwargs.setdefault("col", col)
    facet = atlas[name].plot(**kwargs)
    set_real_aspect(facet.axes)
    # automatically adjust size
    ratio = facet.axes.flatten()[0].get_data_ratio()
    num = atlas[name].sizes[col]
    col_wrap = kwargs.get("col_wrap", 1)
    facet.fig.set_size_inches((1 * num / col_wrap, 1 * ratio * col_wrap))
    facet.set_titles("{value}")
    return facet


def plot_along_grains(grain_maps: xr.DataArray, col: str = "grain", **kwargs) -> xr.plot.FacetGrid:
    """Plot the grain maps from the grain maps."""
    kwargs.setdefault("col", col)
    facet = grain_maps.plot(**kwargs)
    set_real_aspect(facet.axes)
    # automatically adjust size
    ratio = facet.axes.flatten()[0].get_data_ratio()
    num = grain_maps.sizes[col]
    col_wrap = kwargs.get("col_wrap", 1)
    facet.fig.set_size_inches((1 * num / col_wrap, 1 * ratio * col_wrap))
    facet.set_titles("{value}")
    return facet


def set_real_aspect(axes: typing.Union[plt.Axes, typing.Iterable[plt.Axes]]) -> None:
    """Change all axes to be equal aspect."""
    if isinstance(axes, typing.Iterable):
        for ax in axes:
            set_real_aspect(ax)
    else:
        axes.set_aspect(aspect="equal", adjustable="box")
    return


def invert_yaxis(axes: typing.Union[plt.Axes, typing.Iterable[plt.Axes]]) -> None:
    """Change all axes to be equal aspect."""
    if isinstance(axes, typing.Iterable):
        for ax in axes:
            invert_yaxis(ax)
    else:
        axes.invert_yaxis()
    return


def pixel_to_Q(d1: np.ndarray, d2: np.ndarray, ai: AzimuthalIntegrator) -> xr.DataArray:
    """Map pixel position (d1, d2) to Q in nm-1."""
    arr = xr.DataArray(ai.qCornerFunct(d1, d2))
    arr.attrs["standard_name"] = "Q"
    arr.attrs["units"] = "nm$^{-1}$"
    return arr


def assign_Q_to_atlas(atlas: xr.Dataset, ai: AzimuthalIntegrator) -> xr.Dataset:
    """Assign Q grid to atlas"""
    q = pixel_to_Q(atlas["y"].values, atlas["x"].values, ai)
    dims = atlas["y"].dims
    return atlas.assign({"Q": (dims, q)})


def create_atlas_dask(frames: xr.DataArray, windows: pd.DataFrame, verbose: bool = False) -> xr.DataArray:
    """Create a list of tasks to compute the grain maps. Each task is one grain map.

    The dataframe has columns "x", "y", "dx", "dy". Each row is a window on the frames. The window is
    (x - dx, x + dx) in horizontal and (y - dy, y + dy) in vertical. The return is a list of dask arrays.

    Parameters
    ----------
    windows :
        The dataframe with columns "x", "y", "dx", "dy".
    frames :
        One frame.
    verbose :
        Whether to report status or not.

    Returns
    -------
    tasks :
        A list of dask arrays.
    """
    # make the numbers ints
    windows = windows[["x", "y", "dx", "dy"]].apply(np.round).applymap(int)
    # get limits
    nf, _, ny, nx = frames.shape
    # create tasks
    maps = []
    for i, frame in enumerate(frames):
        if verbose:
            print("process frame {}/{}.".format(i + 1, nf), end="\r")
        frame = frame.compute()
        mean_frames = []
        for row in windows.itertuples():
            slice_y = slice(max(row.y - row.dy, 0), min(row.y + row.dy, ny))
            slice_x = slice(max(row.x - row.dx, 0), min(row.x + row.dx, nx))
            mean_frame = frame[:, slice_y, slice_x].mean()
            mean_frames.append(mean_frame)
        maps.append(xr.concat(mean_frames, dim="grain"))
    return xr.concat(maps, dim="time")


def create_dataset(maps: xr.DataArray, windows: pd.DataFrame, metadata: dict, inverted: bool = True) -> xr.Dataset:
    """Create the dataset from the results of create_atlas_dask."""
    ds: xr.Dataset = windows.to_xarray()
    old_dims = list(ds.dims)
    new_values = reshape_to_matrix(maps.values, metadata)
    new_coords = _get_coords(metadata, inverted=inverted)
    new_dims = list(new_coords.keys()) + old_dims
    ds = ds.assign_coords(new_coords)
    ds = ds.assign({"maps": (new_dims, new_values)})
    ds = ds.rename({old_dims[0]: "grain"})
    return ds


def reshape_to_matrix(arr: np.ndarray, metadata: dict) -> np.ndarray:
    if "shape" not in metadata:
        raise KeyError("Missing key '{}' in metadata.".format("shape"))
    reshaped = np.apply_along_axis(
        lambda x: _reshape(x, metadata["shape"], metadata.get("snaking", [])),
        0,
        arr
    )
    return reshaped


def min_and_max_along_time(data: xr.DataArray) -> xr.DataArray:
    """Extract the minimum and maximum values of each pixel in a series of mean frames. Return a data array.
    First is the min array and the second is the max array."""
    min_arr = max_arr = np.mean(data[0].values, axis=0)
    num = data.shape[0]
    my_print("Process frame: {} / {}".format(1, num), end="\r")
    for i in range(1, len(data)):
        arr = np.mean(data[i].values, axis=0)
        min_arr = np.fmin(min_arr, arr)
        max_arr = np.fmax(max_arr, arr)
        my_print("Process frame: {} / {}".format(i + 1, num), end="\r")
    return xr.DataArray(np.stack([min_arr, max_arr]), coords={"dim_0": ["dark", "light"]})


def reshape_to_xarray(arr: xr.DataArray, metadata: dict) -> xr.DataArray:
    """Reshape a 1D data array to an 2D data array according to the metadata about the bluesky plan.
    The coordinates will also be reshaped."""
    data = reshape_to_matrix(arr.values, metadata)
    coords = {name: xr.DataArray(reshape_to_matrix(coord.values, metadata)) for name, coord in arr.coords.items()}
    return xr.DataArray(data, coords=coords)


def reformat_data(ds: xr.Dataset) -> xr.Dataset:
    """Reformat the data loaded from the databroker."""
    frames = np.arange(0, ds["time"].shape[0], 1)
    return ds.rename_dims({"time": "frame"}).reset_index("time").drop_vars("time_").assign_coords(
        {"frame": frames})


def draw_windows(df: pd.DataFrame, ax: plt.Axes) -> None:
    """Draw windows on the axes."""
    for row in df.itertuples():
        xy = (row.x - row.dx - 0.5, row.y - row.dy - 0.5)
        width = row.dx * 2 + 1
        height = row.dy * 2 + 1
        patch = patches.Rectangle(xy, width=width, height=height, linewidth=2, edgecolor="r", fill=False)
        ax.add_patch(patch)
    return


def create_windows_from_size(df: pd.DataFrame, multiplier: float) -> pd.DataFrame:
    df2 = pd.DataFrame()
    df2["x"] = df["x"].apply(np.round).apply(int)
    df2["y"] = df["y"].apply(np.round).apply(int)
    df2["dx"] = df2["dy"] = (df["size"] * multiplier).apply(np.round).apply(int)
    return df2


def create_windows_from_width(df: pd.DataFrame, width: int) -> pd.DataFrame:
    df2 = pd.DataFrame()
    df2["x"] = df["x"].apply(np.round).apply(int)
    df2["y"] = df["y"].apply(np.round).apply(int)
    df2["dx"] = df2["dy"] = int(np.round(width))
    return df2


def track_peaks(frames: xr.DataArray, windows: pd.DataFrame) -> xr.DataArray:
    """Create a list of tasks to compute the grain maps. Each task is one grain map."""
    # get limits
    nf, _, ny, nx = frames.shape
    # create tasks
    maps = []
    for i, frame in enumerate(frames):
        my_print("Process frame {} / {}.".format(i + 1, nf), end="\r")
        frame = frame.compute()
        mean_frames = []
        for row in windows.itertuples():
            slice_y = slice(max(row.y - row.dy, 0), min(row.y + row.dy, ny))
            slice_x = slice(max(row.x - row.dx, 0), min(row.x + row.dx, nx))
            mean_frame = frame[:, slice_y, slice_x].mean()
            mean_frames.append(mean_frame)
        maps.append(xr.concat(mean_frames, dim="grain"))
    return xr.concat(maps, dim="time")


def create_grain_maps(frames: xr.DataArray, windows: pd.DataFrame, metadata: dict,
                      inverted: bool = True) -> xr.Dataset:
    """Create grain maps from frames.

    Parameters
    ----------
    frames :
        The dask array of raw diffraction frames.
    windows :
        The data frame of x, y, dx, dy. The x, y positions in pixels and the half width of the window.
    metadata :
        The start document of the run.
    inverted :
        Use x[::-1] and y[::-1] in grain maps.

    Returns
    -------
    ds :
        The dataset of grain maps.
    """
    maps = track_peaks(frames, windows)
    return create_dataset(maps, windows, metadata, inverted=inverted)


def select_frames(
    image_sum_data: xr.DataArray, metadata: dict, start_row: int = 0, end_row: int = None, **kwargs
) -> None:
    """Select the frames according to the summation of the intensity on the image."""
    image_sum_matrix = reshape_to_xarray(image_sum_data, metadata)
    kwargs.setdefault("size", 8)
    if end_row is None:
        end_row = image_sum_matrix.shape[0]
    image_sum_matrix = image_sum_matrix[start_row:end_row]
    facet = image_sum_matrix.plot(**kwargs)
    index = image_sum_matrix.frame.values
    start_index, end_index = index.min(), index.max()
    facet.axes.set_title("From frame {} to frame {}".format(start_index, end_index))
    set_real_aspect(facet.axes)


def average_intensity(frame: np.ndarray, windows: pd.DataFrame) -> np.ndarray:
    """Calculate the average intensity in windows. Return an array of average intensity."""
    if frame.ndim < 2:
        raise CalculatorError("frame must have no less than 2 dimensions.")
    elif frame.ndim > 2:
        n = frame.ndim - 2
        frame = frame.mean(axis=tuple(range(n)))
    ny, nx = frame.shape
    # create tasks
    I_in_windows = []
    for row in windows.itertuples():
        slice_y = slice(max(row.y - row.dy, 0), min(row.y + row.dy + 1, ny))
        slice_x = slice(max(row.x - row.dx, 0), min(row.x + row.dx + 1, nx))
        I_in_window = np.mean(frame[slice_y, slice_x], dtype=frame.dtype)
        I_in_windows.append(I_in_window)
    return np.array(I_in_windows)


def track_peaks2(frames: xr.DataArray, windows: pd.DataFrame) -> np.ndarray:
    """Create a list of tasks to compute the grain maps. Each task is one grain map."""
    # create intensity vs time for each grain
    global _VERBOSE
    intensities = []
    n = frames.shape[0]
    iter_range = range(n)
    if _VERBOSE > 0:
        iter_range = tqdm.tqdm(iter_range)
    for i in iter_range:
        intensity = average_intensity(frames[i].values, windows)
        intensities.append(intensity)
    # axis: grain, frame
    return np.stack(intensities).transpose()


def summary_in_a_dataset(arr: xr.DataArray, windows: pd.DataFrame):
    # attach the arr to the data set
    ds: xr.Dataset = windows.reset_index(drop=True).to_xarray()
    dims = list(ds.dims.keys())
    ds = ds.rename({dims[0]: "grain"}).assign({"maps": arr}).transpose("grain", "frame")
    return ds


def reshape_to_ndarray(arr: np.ndarray, metadata: dict) -> np.ndarray:
    if "shape" not in metadata:
        raise CalculatorError("Missing key '{}' in metadata.".format("shape"))
    shape = list(arr.shape)[:-1]
    shape.extend(metadata["shape"])
    arr: np.ndarray = arr.reshape(shape)
    # if snaking the row
    if "snaking" in metadata and len(metadata["snaking"]) > 1 and metadata["snaking"][1]:
        if len(metadata["shape"]) != 2:
            raise CalculatorError("snaking only works for the 2 dimension array.")
        n = arr.shape[1]
        for i in range(n):
            if i % 2 == 1:
                arr[:, i, :] = arr[:, i, ::-1]
    return arr


def reshape_to_xarray2(arr: xr.DataArray, metadata: dict) -> xr.DataArray:
    """Reshape a 1D data array to an 2D data array according to the metadata about the bluesky plan.
    The coordinates will also be reshaped."""
    if "shape" not in metadata:
        raise KeyError("Missing 'shape' in metadata.")
    if arr.ndim != 2:
        raise ValueError("The arr must have 2 dimensions. This has {}.".format(arr.ndim))
    shape = metadata["shape"]
    dims = [arr.dims[0]]
    dims.extend(["dim_{}".format(i + 1) for i in range(len(metadata["shape"]))])
    data = xr.DataArray(reshape_to_ndarray(arr.values.copy(), metadata), dims=dims)
    if "extents" in metadata:
        extents = metadata["extents"]
        coords = {"dim_{}".format(i + 1): np.linspace(*extent, num)
                  for i, (extent, num) in enumerate(zip(extents, shape))}
        data = data.assign_coords(coords)
    return data


def create_windows_from_width2(df: pd.DataFrame, width: int) -> pd.DataFrame:
    width = int(width)
    df2 = pd.DataFrame()
    df2["y"] = df["y"].apply(math.floor)
    df2["dy"] = width
    df2["x"] = df["x"].apply(math.floor)
    df2["dx"] = width
    return df2


def min_and_max_along_time2(data: xr.DataArray) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Extract the minimum and maximum values of each pixel in a series of mean frames. Return a data array.
    First is the min array and the second is the max array."""
    global _VERBOSE
    min_arr = max_arr = np.mean(data[0].values, axis=0)
    iter_range = range(1, len(data))
    if _VERBOSE > 0:
        iter_range = tqdm.tqdm(iter_range)
    for i in iter_range:
        arr = np.mean(data[i].values, axis=0)
        min_arr = np.fmin(min_arr, arr)
        max_arr = np.fmax(max_arr, arr)
    return min_arr, max_arr


def get_coords2(start_doc: dict) -> typing.List[np.ndarray]:
    """Get coordinates."""
    if "shape" not in start_doc:
        raise KeyError("Missing key '{}' in the metadata.".format("shape"))
    if "extents" not in start_doc:
        raise KeyError("Missing key '{}' in the metadata".format("extents"))
    shape = start_doc["shape"]
    extents = [np.asarray(extent) for extent in start_doc["extents"]]
    return [np.linspace(*extent, num) for extent, num in zip(extents, shape)]


def show_npy_array(template: str, index: int, **kwargs):
    f = template.format(index)
    arr = xr.DataArray(np.load(f))
    ax = plt.gca()
    arr.plot.imshow(**kwargs, ax=ax)
    ax.set_title(f)
    set_real_aspect(ax)
    plt.show()
    return


def show_tiff_array(template: str, index: int, **kwargs):
    f = template.format(index)
    arr = xr.DataArray(fabio.openimage.open(f).data)
    ax = plt.gca()
    arr.plot.imshow(**kwargs, ax=ax)
    ax.set_title(f)
    set_real_aspect(ax)
    plt.show()
    return


def limit_3std(da: xr.DataArray) -> typing.Tuple[float, float]:
    """Return the mean - 3 * std and mean + 3 * std of the data array.

    Parameters
    ----------
    da

    Returns
    -------

    """
    m, s = da.mean(), da.std()
    return m - 3 * s, m + 3 * s


def plot_crystal_maps(
    da: xr.DataArray,
    limit_func: typing.Callable = None,
    invert_y: bool = False,
    **kwargs
) -> FacetGrid:
    """Plot the crystal maps.

    Parameters
    ----------
    da
    limit_func
    invert_y
    kwargs

    Returns
    -------

    """
    if limit_func is None:
        limit_func = limit_3std
    kwargs.setdefault("col", "grain")
    kwargs.setdefault("col_wrap", 20)
    kwargs.setdefault("sharex", False)
    kwargs.setdefault("sharey", False)
    kwargs.setdefault("add_colorbar", False)
    vmin, vmax = limit_func(da)
    kwargs.setdefault("vmax", vmax)
    kwargs.setdefault("vmin", vmin)
    facet = da.plot.imshow(**kwargs)
    set_real_aspect(facet.axes)
    if invert_y:
        if kwargs.get("sharey"):
            # if y is shared, only the first one need to be inverted
            invert_yaxis(facet.axes.flatten()[0])
        else:
            invert_yaxis(facet.axes)
    return facet


def plot_rocking_curves(da: xr.DataArray, **kwargs) -> FacetGrid:
    """Plot the rocking curves.

    Parameters
    ----------
    da
    kwargs

    Returns
    -------

    """
    kwargs.setdefault("col", "grain")
    kwargs.setdefault("col_wrap", 5)
    kwargs.setdefault("sharex", False)
    kwargs.setdefault("sharey", False)
    return da.plot.line(**kwargs)


def auto_plot(
    da: xr.DataArray,
    title: typing.Tuple[str, str] = None,
    invert_y: bool = False,
    **kwargs
) -> FacetGrid:
    """Automatically detect the data type and plot the data array.

    Parameters
    ----------
    da :
        The data array containing the intensity.
    title :
        Determine title of the axes. It is a format string and a name of the variable. If None, do nothing.
    invert_y :
        Invert the y axis.
    kwargs :
        The kwargs for the configuration of the plot.

    Returns
    -------
    Usually a FacetGrid object.
    """
    if da.ndim <= 1:
        facet = da.plot(**kwargs)
    elif da.ndim == 2:
        facet = plot_rocking_curves(da, **kwargs)
    elif da.ndim == 3:
        facet = plot_crystal_maps(da, invert_y=invert_y, **kwargs)
    else:
        kwargs.setdefault("col", "grain")
        facet = da.plot(**kwargs)
    if title is not None:
        v_name, f_title = title
        vals: np.ndarray = da[v_name].values
        axes: typing.List[plt.Axes] = facet.axes.flatten()
        for ax, val in zip(axes, vals):
            ax.set_title(f_title.format(val))
    return facet


def auto_plot_dataset(
    ds: xr.Dataset,
    key: str = "intensity",
    title: typing.Tuple[str, str] = None,
    invert_y: bool = False,
    **kwargs
):
    facet = auto_plot(ds[key], title=None, invert_y=invert_y, **kwargs)
    if title is not None:
        v_name, f_title = title
        vals: np.ndarray = ds[v_name].values
        axes: typing.List[plt.Axes] = facet.axes.flatten()
        for ax, val in zip(axes, vals):
            ax.set_title(f_title.format(val))
    return facet


def set_vlim(kwargs: dict, da: xr.DataArray, alpha: float, low_lim: float = 0.,
             high_lim: float = float("inf")) -> None:
    mean = da.values.mean()
    std = da.values.std()
    kwargs.setdefault("vmin", max(low_lim, mean - std * alpha))
    kwargs.setdefault("vmax", min(high_lim, mean + std * alpha))
    return


class CalculatorError(Exception):
    pass


class Calculator(object):
    """The Calculator of the crystal maps.

    Each of the attribute can be the calculated output and the input for the next step of calculation.

    Attributes
    ----------
    frames_arr : DataArray
        A data array of diffraction data. It is assumed to be a (N, F, Y, X) shape array, where N is the number
        of exposure, F is the number of the frames in one exposure, the Y is the number of pixels in vertical,
        X is the number of pixels in horizontal. It can also be (N, Y, X) shape array if there is only one
        frame in each exposure.
    dark : ndarray
        The minimum values on the pixels in whole series of exposures. It is assumed to be the background.
    light: ndarray
        The maximum values on the pixels in the whole series of exposures. It is assumed to be a image of all
        the Bragg peaks on the background.
    peaks: DataFrame
        A dataframe records all the peak positions.
    windows: DataFrame
        A dataframe records all the center positions and half width of the windows.
    intensity: ndarray
        The array of the average intensity in each of the windows in the series of exposures. It has the shape (
        W, dim_0, dim_1, ...). The W is the number of the windows and the dim_i is the length of the dimension
        i in the grid scan.
    bkg_intensity : ndarray
        The integrated background intensity in the windows.
    coords : ndarray
        The coordinates of the dim_0, dim_1, ... in grid scan.
    metadata : dict
        The dictionary to record the metadata. There are two important keys: shape, extents. The shape is the
        number of points in each dimension in the grid scan. It should be a array like (dim_0, dim_1, ...). The
        extents are the start and end value for each dimension. It should be array like [(start0, end0),
        (start1, end1), ...]. These two keys are used to calculate the coordinates.
    ai : AzimuthalIntegrator
        The class to hold the data of the geometry of the experiments. It is used to index the Q values for each
        window.
    """

    def __init__(self):
        # dims: time, ..., pixel_y, pixel_x
        self.frames_arr: typing.Union[None, xr.DataArray] = None
        # dims: pixel_y, pixel_x
        self.dark: typing.Union[None, np.ndarray] = None
        # dims: pixel_y, pixel_x
        self.light: typing.Union[None, np.ndarray] = None
        # index: all_grain
        self.peaks: typing.Union[None, pd.DataFrame] = None
        # index: grain
        self.windows: typing.Union[None, pd.DataFrame] = None
        # dims: grain, dim_1, ..., dim_n
        self.intensity: typing.Union[None, np.ndarray] = None
        # dims: grain
        self.bkg_intensity: typing.Union[None, np.ndarray] = None
        # dims: dim_1, ..., dim_n
        self.coords: typing.Union[None, typing.List[np.ndarray]] = None
        # keys: shape, extents, snaking
        self.metadata: typing.Union[None, dict] = None
        # pyFAI
        self.ai: typing.Union[None, AzimuthalIntegrator] = None

    def _check_attr(self, name: str):
        if getattr(self, name) is None:
            raise CalculatorError("Attribute '{}' is None. Please set it.".format(name))
        if name == "metadata" and "shape" not in self.metadata:
            raise CalculatorError("There is no key 'shape' in the metadata.")

    def squeeze_shape_and_extents(self) -> None:
        """Squeeze the shape and extents so that it only has the dimension with length > 1."""
        self._check_attr("metadata")
        shape = self.metadata["shape"]
        n = len(shape)
        # get the index of the dimension > 1
        index = {i for i in range(n) if shape[i] > 1}
        self.metadata["shape"] = [shape[i] for i in range(n) if i in index]
        if "extents" in self.metadata:
            extents = self.metadata["extents"]
            self.metadata["extents"] = [extents[i] for i in range(n) if i in index]
        return

    def calc_dark_and_light_from_frames_arr(self, index_range: slice = None):
        """Get the light and dark frame in a series of frames."""
        self._check_attr("frames_arr")
        frames_arr = self.frames_arr[index_range] if index_range is not None else self.frames_arr
        self.dark, self.light = min_and_max_along_time2(frames_arr)

    def calc_peaks_from_light_frame(self, diameter: typing.Union[int, tuple], *args, **kwargs):
        """Get the Bragg peaks on the light frame."""
        self._check_attr("light")
        self.peaks = tp.locate(self.light, diameter, *args, **kwargs)

    def calc_peaks_from_dk_sub_frame(self, diameter: typing.Union[int, tuple], *args, **kwargs):
        """Get the Bragg peaks on the light frame."""
        self._check_attr("light")
        light = self.light if self.dark is None else self.light - self.dark
        self.peaks = tp.locate(light, diameter, *args, **kwargs)

    def calc_windows_from_peaks(self, num: int, width: int):
        """Gte the windows for the most brightest Bragg peaks."""
        self._check_attr("peaks")
        df = self.peaks.nlargest(num, "mass")
        self.windows = create_windows_from_width2(df, width)

    def calc_intensity_in_windows(self):
        """Get the intensity array as a function of index of frames."""
        self._check_attr("frames_arr")
        self._check_attr("windows")
        self.intensity = track_peaks2(self.frames_arr, self.windows)
        if self.dark is not None:
            self.bkg_intensity = average_intensity(self.dark, self.windows)
            self.intensity = (self.intensity.transpose() - self.bkg_intensity).transpose()
        else:
            my_print("Attribute 'dark' is None. No background correction.")

    def assign_q_values(self):
        """Assign the values to the windows dataframe."""
        self._check_attr("ai")
        self._check_attr("windows")
        self.windows["Q"] = self.ai.qFunction(self.windows["x"], self.windows["y"])

    def reshape_intensity(self):
        """Reshape the intensity array."""
        self._check_attr("metadata")
        self._check_attr("intensity")
        self.intensity = reshape_to_ndarray(self.intensity, self.metadata)

    def calc_coords(self):
        """Calculate the coordinates."""
        self._check_attr("metadata")
        self.coords = get_coords2(self.metadata)

    def dark_to_xarray(self) -> xr.DataArray:
        """Convert the dark image to DataArray."""
        return xr.DataArray(self.dark, dims=["pixel_y", "pixel_x"])

    def light_to_xarray(self) -> xr.DataArray:
        """Convert the light image to DataArray"""
        return xr.DataArray(self.light, dims=["pixel_y", "pixel_x"])

    def intensity_to_xarray(self) -> xr.DataArray:
        """Convert the intensity array to DataArray"""
        dims = ["dim_{}".format(i) for i in range(self.intensity.ndim - 1)]
        arr = xr.DataArray(self.intensity, dims=["grain"] + dims)
        if self.coords is not None:
            coords = self.coords_to_dict()
            arr = arr.assign_coords(coords)
        return arr

    def windows_to_xarray(self) -> xr.DataArray:
        """Convert the windows DataFrame to xarray."""
        return self.windows.rename_axis(index="grain").to_xarray()

    def coords_to_dict(self) -> dict:
        """Convert the coordinates to dictionary."""
        return {"dim_{}".format(i): coord for i, coord in enumerate(self.coords)}

    def to_dataset(self) -> xr.Dataset:
        """Convert the calculation results to DataSet."""
        dct = dict()
        if self.dark is not None:
            dct["dark"] = self.dark_to_xarray()
        if self.light is not None:
            dct["light"] = self.light_to_xarray()
        if self.intensity is not None:
            dct["intensity"] = self.intensity_to_xarray()
        ds = xr.Dataset(dct)
        if self.windows is not None:
            ds2 = self.windows_to_xarray()
            ds = ds.merge(ds2)
        if self.metadata is not None:
            ds = ds.assign_attrs(**self.metadata)
        return ds

    def get_frame(self, index: int) -> xr.DataArray:
        """Get the frame of at the index."""
        self._check_attr("frames_arr")
        return self.frames_arr[index].compute().mean(axis=0)

    def show_frame(self, index: int, *args, **kwargs) -> FacetGrid:
        """Show the frame at that index."""
        frame = self.get_frame(index)
        set_vlim(kwargs, frame, 4.)
        facet = frame.plot.imshow(*args, **kwargs)
        set_real_aspect(facet.axes)
        return facet

    def show_windows_on_frame(self, index: int, *args, **kwargs) -> FacetGrid:
        """Show the windows on the frame at the index."""
        self._check_attr("windows")
        facet = self.show_frame(index, *args, **kwargs)
        draw_windows(self.windows, facet.axes)
        return facet

    def show_dark(self, *args, **kwargs) -> FacetGrid:
        """Show the dark image."""
        self._check_attr("dark")
        frame = self.dark_to_xarray()
        set_vlim(kwargs, frame, 4.)
        facet = frame.plot.imshow(*args, **kwargs)
        set_real_aspect(facet.axes)
        return facet

    def show_light(self, *args, **kwargs) -> FacetGrid:
        """Show the light image."""
        self._check_attr("light")
        frame = self.light_to_xarray()
        set_vlim(kwargs, frame, 4.)
        facet = frame.plot.imshow(*args, **kwargs)
        set_real_aspect(facet.axes)
        return facet

    def show_light_sub_dark(self, *args, **kwargs) -> FacetGrid:
        """Show the dark subtracted light image."""
        self._check_attr("light")
        light = self.light_to_xarray()
        if self.dark is not None:
            dark = self.dark_to_xarray()
            light = np.subtract(light, dark)
        set_vlim(kwargs, light, 4.)
        facet = light.plot.imshow(*args, **kwargs)
        set_real_aspect(facet.axes)
        return facet

    def show_windows(self, *args, **kwargs) -> FacetGrid:
        """Show the windows on the dark subtracted light image."""
        self._check_attr("light")
        self._check_attr("windows")
        facet = self.show_light_sub_dark(*args, **kwargs)
        draw_windows(self.windows, facet.axes)
        return facet

    def show_intensity(self, **kwargs) -> FacetGrid:
        """Show the intensity array."""
        arr = self.intensity_to_xarray()
        return auto_plot(arr, **kwargs)

    @staticmethod
    def auto_visualize(
        ds: xr.Dataset,
        key: str = "intensity",
        title: typing.Tuple[str, str] = None,
        invert_y: bool = False,
        **kwargs
    ) -> FacetGrid:
        """Automatically plot the intensity array in the dataset."""
        return auto_plot_dataset(ds, key, title, invert_y, **kwargs)

    def auto_process(self, num_wins: int, hw_wins: int, diameter: int, index_filter: slice = None,
                     *args, **kwargs) -> None:
        """Automatically process the data in the standard protocol.

        Parameters
        ----------
        num_wins : int
            The number of windows.
        hw_wins : int
            The half width of the windows in pixels.
        diameter : int
            The diameter of the kernel to use in peak finding in pixels. It must be an odd integer.
        index_filter : slice
            The index slice of the data to use in the calculation of the dark and light image.
        args :
            The position arguments of the peak finding function `trackpy.locate`.
        kwargs :
            The keyword arguments of the peak finding function `trackpy.locate`.

        Returns
        -------
        None. The calculation results are saved in attributes.
        """
        if diameter % 2 == 0:
            raise CalculatorError("Diamter must be an odd integer.")
        self.squeeze_shape_and_extents()
        self.calc_dark_and_light_from_frames_arr(index_filter)
        self.calc_peaks_from_dk_sub_frame(diameter, *args, **kwargs)
        self.calc_windows_from_peaks(num_wins, hw_wins)
        self.calc_intensity_in_windows()
        try:
            self.calc_coords()
        except CalculatorError as e:
            print(e)
        try:
            self.assign_q_values()
        except CalculatorError as e:
            print(e)
        try:
            self.reshape_intensity()
        except CalculatorError as e:
            print(e)
        return



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr

import tomography.utils as utils

plt.ioff()


def test_reshape():
    """Test reshape in a snaking axes case."""
    ds = xr.Dataset({"array_data": [0, 0, 0, 1, 0, 0, 0, 0, 0]})
    ds.attrs["shape"] = (3, 3)
    ds.attrs["extents"] = [(-1, 1), (-1, 1)]
    ds.attrs["snaking"] = [False, True]
    res = utils.reshape(ds, "array_data")
    expected = np.zeros((3, 3))
    expected[1, 2] = 1
    assert np.array_equal(res.values, expected)


def test_plot_real_aspect():
    """Test plot_real_aspect on a zero image."""
    data = xr.DataArray(np.zeros((5, 5)))
    data[2, 2] = 1
    utils.plot_real_aspect(data)
    # plt.show()
    plt.clf()


def test_annotate_peaks():
    """Test annotate_peaks in a toy case."""
    df = pd.DataFrame(
        {"y": [2], "x": [2], "mass": [1], "size": [1], "ecc": [1], "signal": [1], "raw_mass": [1], "frame": [0]}
    )
    image = xr.DataArray(np.zeros((5, 5)))
    image[2, 2] = 1
    utils.annotate_peaks(df, image)
    # plt.show()
    plt.clf()


def test_create_atlas():
    """Test create_atlas by plotting the figure out. Test assign_Q_to_atlas"""
    df = pd.DataFrame(
        {"y": [1, 2], "x": [1, 2], "mass": [1, 2], "size": [1, 2], "ecc": [1, 2], "signal": [1, 2],
         "raw_mass": [1, 2], "frame": [1, 3], "particle": [0, 1]}
    )
    df.attrs["shape"] = (2, 2)
    df.attrs["extents"] = [(-3, 3), (-1, 1)]
    df.attrs["snaking"] = (False, True)
    ds = utils.create_atlas(df)
    # print(ds)
    facet = utils.plot_grain_maps(ds)
    facet.fig.set_size_inches(4, 4)
    # facet.fig.show()
    plt.clf()
    # test
    ai = utils.AzimuthalIntegrator(detector="Perkin", wavelength=2 * np.pi)
    ds2 = utils.assign_Q_to_atlas(ds, ai)
    print(ds2)


def test_map_to_Q():
    """Test map_to_Q_vector."""
    # test numpy
    d1 = np.array([0, 1])
    d2 = np.array([0, 1])
    ai = utils.AzimuthalIntegrator(detector="Perkin", wavelength=2 * np.pi)
    q = utils.pixel_to_Q(d1, d2, ai)
    assert q.shape == (2,)


def test_Calculator_1():
    c = utils.Calculator()

    c.frames_arr = xr.DataArray([[[[1, 0], [0, 0]]], [[[0, 0], [1, 1]]]])
    c.show_frame(0)
    c.show_frame(1)
    c.calc_dark_and_light_from_frames_arr()
    expect0 = np.array([[0, 0], [0, 0]])
    expect1 = np.array([[1, 0], [1, 1]])
    assert np.array_equal(c.dark, expect0)
    assert np.array_equal(c.light, expect1)

    c.show_dark()
    plt.show(block=False)
    plt.clf()
    c.show_light()
    plt.show(block=False)
    plt.clf()

    c.calc_peaks_from_light_frame(1, noise_size=0)
    expect2 = pd.DataFrame(columns=["y", "x", "mass", "size", "ecc", "signal", "raw_mass"])
    assert c.peaks.equals(expect2)

    c.peaks = pd.DataFrame([[1.5, 0.5, 3], [0.5, 0.5, 2], [1.5, 1.5, 1]],
                           columns=["y", "x", "mass"])
    c.calc_windows_from_peaks(100, 0)
    expect3 = pd.DataFrame([[1, 0, 0, 0], [0, 0, 0, 0], [1, 0, 1, 0]], columns=["y", "dy", "x", "dx"])
    assert c.windows.equals(expect3)

    c.show_windows()
    plt.show(block=False)
    plt.clf()

    c.calc_intensity_in_windows()
    expect4 = np.array([[0., 1.], [1., 0], [0., 1.]])
    assert np.array_equal(c.intensity, expect4)

    c.show_intensity()
    plt.show(block=False)
    plt.clf()

    ds = c.to_dataset()
    print(ds)


def test_Calculator_2():
    c = utils.Calculator()

    c.metadata = {"shape": [2, 2], "extents": [(-1, 0), (1, 3)], "snaking": (False, True)}
    c.intensity = np.array([[1, 2, 3, 4], [4, 3, 2, 1]])
    c.reshape_intensity()
    expect5 = np.array([[[1, 2], [4, 3]], [[4, 3], [1, 2]]])
    assert np.array_equal(c.intensity, expect5)

    c.show_intensity()
    plt.show(block=False)
    plt.clf()

    c.calc_coords()
    assert np.array_equal(c.coords[0], np.array([-1., 0.]))
    assert np.array_equal(c.coords[1], np.array([1., 3.]))

    c.show_intensity()
    plt.show(block=False)
    plt.clf()


def test_Calculator_3():
    c = utils.Calculator()
    c.windows = pd.DataFrame([[1, 0, 0, 0], [0, 0, 0, 0], [1, 0, 1, 0]], columns=["y", "dy", "x", "dx"])
    c.ai = utils.AzimuthalIntegrator(detector="Perkin", wavelength=2 * np.pi)
    c.assign_q_values()
    print(c.windows)
